
class ExploitNotOpsecException(Exception):
    """Exception raised when an exploit is not opsec safe for check"""

class BadCredentialsException(Exception):
    """Exception raised when the credentials are not valid"""

class MissingMFAException(Exception):
    """Exception when MFA is needed but not provided"""
